

module RefacCleanImps(cleanImports,mkImpExplicit,addToExport,rmFromExport) where

import PosSyntax
import Maybe
import TypedIds
import List  
import RefacUtils 
import UniqueNames 
import HsName hiding (ModuleName)
import SourceNames
import PNT 
import RefacMoveDef(liftingInClientMod)
import PFE0 (findFile)
import PrettyPrint hiding (group)
import NamesEntities
import Ents 
import Relations
import Modules(mImp',mExpListEntry)
import AST4ModSys (toImport',toExpListEntry') 
import HsIdent(HsIdentI(..))



{- This refactoring trys to remove the unnecessary import declarations from the module-}

cleanImports args
 = do let fileName= ghead "filename" args  
      modName <- RefacUtils.moduleName fileName
      (inscps, _, mod, toks) <- parseSourceFile fileName      
      let inscps' =relToList inscps    
      imp_Ents_List <- getImportedEnts inscps' (hsModImports mod)
      (mod',((toks',_),_))<-runStateT (doCleanImports imp_Ents_List (inscps', mod))
                                        ((toks,unmodified),(fileName,initValue)) 
      writeRefactoredFiles [((fileName,True),(toks', mod'))]


{- This function is not used yet.

 filterOutPreludeEnts inscps
 = let qualifiedEnts = filter (\e ->case e of
                                     (SN (Qual (PlainModule "Prelude") _) _, _) ->True
                                     _ -> False) inscps
   in inscps \\ (qualifiedEnts ++ map unqualify qualifiedEnts)
-}

doCleanImports imp_Ents_List (inscps, mod@(HsModule _ modName exps imps decls))
 = do  -- imp_usedEnts_List ::[(ImportDeclP, [PNames])]
      let imp_usedEnts_List =map (getUsedEnts mod) imp_Ents_List
      mod' <- checkEachImport mod (imp_usedEnts_List, imp_usedEnts_List)
      combineImports mod' (hsModImports mod')   

-- remove an import declaration from the module.
rmOneImport mod@(HsModule loc modName exps imps decls) imp
  = do ((toks,_),others)<-get  
       let (startPos,endPos)=startEndLocIncComments toks imp  
           toks'=if startPos ==(0,0) then  toks
                                     else deleteToks toks startPos endPos
       put ((toks',modified),others)
       return (HsModule loc modName exps (imps\\[imp]) decls)


checkEachImport mod@(HsModule loc modName exps imps decls) ([],_)
  =return mod

checkEachImport mod@(HsModule loc modName exps imps decls) ((i@(imp,inscpEnts,usedEnts):is), origList)
  = if usedEnts == []   -- this import decl is not used at all.
      then do mod'<- (do m'<-rmOneImport mod imp
                         if inscpEnts==[] && modIsExported (qualifier imp) mod
                           then rmItemFromExport m' ([qualifier imp], [])
                           else return m')           
              checkEachImport mod' (is, (origList \\ [i]))
      else if any (\(x,y,z)->(usedEnts == z  && length inscpEnts >= length y)
                    || (usedEnts \\ z ==[] && length usedEnts <length z)) (origList \\ [i])    
             then do mod'<-rmOneImport mod imp
                     checkEachImport mod' (is,(origList \\[i]))
             else checkEachImport mod (is, origList)


--getImportedEnts::[(QName,Entity)]->[HsImportDeclP]->[(HsImportDeclP,[(QName,Entity)])]
getImportedEnts inscps imps
 = do importedEnts<-mapM getImportedEnts' imps
      return $ zip imps importedEnts
  where  
    --getImportedEnts'::HsImportDeclP->[(QName,Entity)]
 {-   getImportedEnts' imp@(HsImportDecl _ _ qual Nothing h)
       =return (qualifiedEnts ++ (if qual then [] else map unqualify qualifiedEnts))
        where  qualifiedEnts  = filter (\e->case e of 
                                   ((SN (Qual q _) loc),(Ent modName' _ _))
                                      ->q == qualifier imp 
                                   _  -> False) inscps    -}
    -- This might be ineffienct when there is no module alias.
    getImportedEnts' (imp@(HsImportDecl loc (SN modName _) qual as h)::HsImportDeclP)
       = do fileName <- PFE0.findFile modName  
            (_, exps, mod, toks) <- parseSourceFile fileName       
            return $ relToList (mImp' (listToRel exps) 
                     (toImport' (HsImportDecl loc modName qual (as' as) h)))
      where as' Nothing = Nothing
            as' (Just (SN modName _)) = Just modName
        
getUsedEnts mod@(HsModule _ _ exps imps decls) (imp@(HsImportDecl _ (SN modName _) _ as h),inscpEnts)
  = if modIsExported (qualifier imp) mod
      then (imp, inscpEnts, inscpEnts)
      else case h of
              (Just (False, ents)) ->if ents == [] 
                                       then (imp, inscpEnts,[])
                                       else (imp, inscpEnts, usedEnts' (exps, decls))
              _ -> (imp, inscpEnts, usedEnts' (exps, decls))          
   where 
     usedEnts'= (nub.ghead "getUsedEnts'"). applyTU (full_tdTU (constTU [] `adhocTU` inPnt)) 
        where
          inPnt pnt@(PNT pname ty loc)
           | isTopLevelPNT pnt
            =let ent'=pntToEnt pnt
             in if isJust (find (==ent') inscpEnts)
                  then return [ent']
                  else return []
          inPnt _ = return []


modIsExported modName mod
 = if isNothing exps
     then False
     else isJust $ find (\x->getModName x == Just modName) (fromJust exps)
   where
     exps =  hsModExports mod
     getModName (ModuleE  (SN modName _)) = Just modName
     getModName _ = Nothing
 

unqualify ((SN (Qual _ id) l), ent) = ((SN (UnQual id) l), ent)
unqualify x  = x  

qualifier (HsImportDecl _ (SN modName _) qual Nothing h) = modName
qualifier (HsImportDecl _ _ qual (Just (SN modName _)) h) = modName

combineImports mod [] = return mod  
combineImports mod@(HsModule _ _ _ imps _) (i:is)
   = case i of 
     (HsImportDecl _ modName  qual as (Just(False, ents)))
        ->let sub_is= filter (\x-> case x of
                                  (HsImportDecl _ modName1 qual1 as1 (Just(False,ents1)))
                                    ->modName1 == modName && qual==qual1 && as == as1
                                  _ ->False ) is
          in if sub_is==[] then combineImports mod is
                           else do let ents=concatMap importedEnts sub_is
                                   mod'<-foldM rmOneImport mod sub_is
                                   let (imps1, imps2) = break (==i) imps   
                                   i' <-addEntsToImport i ents 
                                   combineImports (mod' {hsModImports =((imps1++[i']++(tail imps2)) \\ sub_is)}) (is \\ sub_is) 
     _  ->combineImports mod is
  where
     importedEnts (HsImportDecl _ _ False Nothing (Just(False,ents))) =ents
     importedEnts  _ =[] 
    
     addEntsToImport imp@(HsImportDecl l m False Nothing (Just(False,ents))) ents'
       = do  ((toks,_),others)<-get 
             let (_,endPos)=getStartEndLoc toks (glast "addEntsToImport" ents)
                 (t, (_,s))=ghead "addEntsToImport" $ getToks (endPos,endPos) toks
                 newToken=mkToken t endPos (s++","++  showEnts (render.ppi) ents')
                 toks'=replaceToks toks endPos endPos [newToken]
             put ((toks',modified),others)               
             return (HsImportDecl l m False Nothing (Just(False,ents++ents')))
   
     addEntsToImport imp ents'  = return imp

showEnts f [] = ""
showEnts f (ent:ents) = if ents==[] then f ent
                                else (f ent ++ "," ++ showEnts f ents)
    
pntToEnt (PNT pn@(PN (UnQual s) (G modName s1 loc)) ty _)
  = let t= if (hasNameSpace ty == ValueName) then HsVar else HsCon
    in ((SN (UnQual s) (getSrcLoc loc)), (Ent modName (t (SN s (getSrcLoc loc))) (fmap pIdtoId ty)))

pntToEnt (PNT pn@(PN (Qual modName s) (G modName1 s1 loc)) ty _)
  = let t= if (hasNameSpace ty == ValueName) then HsVar else HsCon
    in ((SN (Qual modName s) (getSrcLoc loc)) , (Ent modName1 (HsVar (SN s (getSrcLoc loc))) (fmap pIdtoId ty)))
          
pIdtoId pn@(PN i orig) = (SN i (pNameToOrig pn))

getSrcLoc (N (Just loc)) = loc
getSrcLoc (N Nothing)    = loc0


-- this refactoring makes the all the used enitities from a import decl explicit in the import declaration.
mkImpExplicit args
 =do let fileName= ghead "filename" args            
         row = read (args!!1)::Int
         col = read (args!!2)::Int
     modName <- RefacUtils.moduleName fileName
     (inscps, _, mod, toks) <- parseSourceFile fileName      
     let imp'=locToImport fileName row col mod   
     unless (isJust imp')
        $ error ("Invalid cursor position! "++
                "Please point the cursor to the beginning position of the module name." )
     let inscps' = relToList inscps  
         imp = fromJust imp'    
     [(imp, importedEnts)] <- getImportedEnts inscps' [imp]  
     (mod',((toks',_),_))<-runStateT (doExplicitImport mod (imp,importedEnts))
                                         ((toks,unmodified),(fileName,initValue))
     writeRefactoredFiles [((fileName,True),(toks', mod'))]


locToImport fileName row col mod
  = applyTU (once_buTU (failTU `adhocTU` worker)) mod
      where
       worker (imp@(HsImportDecl l (SN modName (SrcLoc _ _ row1 col1)) _ _ _)::HsImportDeclP)
         | row1 == row && col1 == col 
         =Just imp
       worker _ = Nothing 
                     
doExplicitImport mod (imp@(HsImportDecl l m qual as _),importedEnts)
 = do let used =usedEnts importedEnts  mod          
          (imps1, imps2) = break (==imp) (hsModImports mod)
          imp' = if (used==[])
                   then (HsImportDecl l m qual as (Just(False, [])))
                   else (HsImportDecl l m qual as (Just (False, used)))
      ((toks,_),others)<-get
      let (startPos, endPos) = getStartEndLoc toks imp
          newToken = mkToken Varid startPos ((render.ppi) imp')
          toks' = replaceToks toks startPos endPos [newToken]
      put ((toks',modified),others)             
      return (mod {hsModImports=(imps1++[imp']++imps2)})

  where
   usedEnts importedEnts = combineEntSpecs. (nub.ghead "usdEnts'"). 
                            applyTU (full_tdTU (constTU [] `adhocTU` inPnt)) 
        where
          inPnt pnt@(PNT pname ty loc)
           | isTopLevelPNT pnt
            =let ent'=pntToEnt pnt
             in if isJust (find (==ent') importedEnts)
                  then return [pNTtoEntSpec pnt]
                  else return []
          inPnt _ = return []
     
   combineEntSpecs [] = []
   combineEntSpecs (e:es) 
       = case e of
          Abs pnt -> if findDup pnt es then combineEntSpecs es else (e:combineEntSpecs es)
               where findDup pnt es = any (\x->case x of 
                                            ListSubs i is -> pNTtoName i== pNTtoName pnt
                                            Abs  i        -> pNTtoName i== pNTtoName pnt
                                            _             -> False) es
          ListSubs p _ -> if findDup p es then combineEntSpecs (doCombine e es)  else (e:combineEntSpecs es)
              where findDup pnt es = any (\x->case x of 
                                              ListSubs p' _ ->pNTtoName p == pNTtoName p' 
                                              _  -> False) es        
                    doCombine (ListSubs p cs) es
                          = let (es1, es2) = break (\x-> case x of
                                                    ListSubs p' _ ->pNTtoName p == pNTtoName p'  
                                                    _ ->False) es 
                                (ListSubs _ cs') = ghead "doCombine" es2
                            in (es1++[ListSubs p (cs++cs')]++tail es2) 
                                
          _  -> (e: combineEntSpecs es)
       
pNTtoEntSpec pnt@(PNT pn@(PN _ (G modName s1 loc)) ty _)
  = case  ty of      
       ConstrOf i _ -> ListSubs (pIdToPNT i) [HsCon pnt']
       Type _       -> Abs pnt' 
       _            -> Var pnt'
  where   
   pIdToPNT (PN id orig) =PNT (PN (UnQual id) orig) Value  (N Nothing) 
 
   pnt'= rmQualifier pnt 
     where rmQualifier (PNT (PN (Qual modName  s) l) ty loc)= (PNT (PN (UnQual s) l) ty loc)
           rmQualifier  x  = x 


-- this refactoring adds an user-selected top-level identifier to the export list.

addToExport args
 =do let fileName= ghead "filename" args            
         row = read (args!!1)::Int
         col = read (args!!2)::Int   
     modName <- RefacUtils.moduleName fileName
     (inscps, exps, mod, toks) <- parseSourceFile fileName
     let pnt = locToPNT fileName row col mod                                            
     unless (pnt/=defaultPNT) $ error "Invalid cursor position!\n"  -- this refactoring does not handle module names.
     unless (isTopLevelPNT pnt) $ error "The selected identifier is not a top level identifier!\n"
     (mod',((toks',_),_))<-runStateT (doAddingToExport pnt mod)
                                         ((toks,unmodified),(fileName,initValue))
     clients <-clientModsAndFiles modName
     refactoredClients <-mapM (refactorInClientMod modName (pNTtoPN pnt)) clients
     writeRefactoredFiles $ ((fileName,True),(toks', mod')):refactoredClients
  where 
     doAddingToExport pnt mod
      = do let pn = pNTtoPN pnt 
           unless (not (isExplictlyExported pn mod  || wholeModIsExported mod))
                   $ error "The indentifier is already exported!\n"
           let exps=fromJust $ hsModExports mod
               newEnt = EntE (pNTtoEntSpec pnt)
           insertEEnt mod newEnt

     insertEEnt mod e@(EntE ent)
      = let exps = fromJust $ hsModExports mod
        in case  ent  of 
            ListSubs i is  ->let (exps1,exps2)=break (\i'->case i' of
                                                         (EntE (ListSubs i'' _)) ->(i''==i)
                                                         (EntE (Abs i''))        ->(i''== i)
                                                         _                       ->False) exps
                             in if exps2 == []
                                 then do addEntInToks mod e 
                                         return $ (mod {hsModExports=Just (exps ++ [e])}) 
                                 else do let e'=ghead "insertEnt" exps2
                                             exps' = case e' of 
                                                     (EntE (ListSubs i'' is')) ->
                                                                Just (exps1++((EntE (ListSubs i (is'++is))):tail exps2))
                                                     (EntE (Abs i''))        ->
                                                                Just (exps1 ++((EntE (ListSubs i'' is)):tail exps2))
                                         addSubEntInToks e' ent
                                         return $ (mod {hsModExports=exps'})
            _              -> do addEntInToks mod e 
                                 return $ (mod {hsModExports=Just (exps ++ [e])})
         
     addEntInToks mod newEnt 
       = do  ((toks, _), others)<-get
             let exps = fromJust $ hsModExports mod
                 (_, (row,col)) =if exps == [] then getStartEndLoc toks (hsModName mod) 
                                               else getStartEndLoc toks (glast "addEntInToks" exps)   
                 (t, (pos,s))   =  gfromJust "addVarItemInImport1" (find isCloseBracket 
                                    (dropWhile (\t->tokenPos t <=(row,col)) toks))
                 sep = if exps == [] then "" else ","
                 newToken =mkToken t (simpPos pos) (sep ++ (render.ppi) newEnt ++s)                            
                 toks' = replaceToks toks (simpPos pos) (simpPos pos) [newToken]
             put ((toks',modified),others)

     addSubEntInToks e subEnt@(ListSubs i (is:_))
       = do ((toks,_), others)<-get
            let toks'=case e of
                        (EntE (ListSubs i' is')) ->
                          let (_, (row,col)) = if is'==[] then getStartEndLoc toks i'
                                                          else getStartEndLoc toks (glast "addEntInToks" is')
                              (t,(pos,s)) = gfromJust "addSubEntInToks" (find isCloseBracket 
                                              (dropWhile (\t->tokenPos t <=(row,col)) toks))
                              newToken= mkToken t (simpPos pos) ("," ++ (render.ppi) is ++s)
                          in replaceToks toks (simpPos pos) (simpPos pos) [newToken]
                        (EntE (Abs i')) ->
                              let (_, (row,col)) = getStartEndLoc toks i' 
                                  (t, (pos,s))   = ghead "addSubEntInToks" (getToks ((row,col),(row,col)) toks)  
                                  newToken= mkToken t (simpPos pos) (s++"("++(render.ppi) is++")")             
                              in replaceToks toks (simpPos pos) (simpPos pos) [newToken]            
            put ((toks',modified),others)
  
     refactorInClientMod serverModName pn (modName, fileName)
       = liftingInClientMod serverModName [pn] (modName, fileName)

--This refactoring removes an user-selected entry from the export list if none of the idenitifiers exported
--by this entry is used by other modules.

rmFromExport args
 =do let fileName= ghead "filename" args            
         row = read (args!!1)::Int
         col = read (args!!2)::Int   
     modName <- RefacUtils.moduleName fileName
     (inscps, exps, mod, toks) <- parseSourceFile fileName
     let eEntry' = locToEEntry row col mod toks             
     unless (eEntry'/=Nothing) $ error "Invalid cursor position!\n"   
     let eEntry = fromJust eEntry'
         ents = exportedEnts inscps eEntry
     if isDuplicatedEntry inscps (eEntry, ents) (fromJust (hsModExports mod))
       then do (mod',((toks',_),_))<-runStateT (doRmFromExport mod eEntry)
                                   ((toks,unmodified),(fileName,initValue)) 
               writeRefactoredFiles [((fileName, modified),(toks',mod'))]
       else do clients<-clientModsAndFiles modName
               refactoredClients <- mapM (refactorInClientMod ents) clients
               (mod',((toks',_),_))<-runStateT (doRmFromExport mod eEntry)
                                        ((toks,unmodified),(fileName,initValue))
               writeRefactoredFiles $ ((fileName,modified),(toks',mod')):refactoredClients
          
     
   where

    -- Calculates all the entities exported by this entry.
    exportedEnts inscps eEntry = relToList (mExpListEntry inscps  (toExpListEntry' eEntry))

    -- Returns True if this entry is duplicated or covered by another entry in the export list.
    isDuplicatedEntry inscps (eEntry, ents) exps 
       =let ents'= map (mExpListEntry' inscps) (map toExpListEntry' (exps \\[eEntry]))        
        in  any (\e->intersect e ents==ents) ents'
        where mExpListEntry' inscps entry= relToList (mExpListEntry inscps entry)
  
    -- Removes an entry from the export list.
    doRmFromExport mod eEntry = rmEntsFromExport mod [eEntry]

    -- get the user selected entry from the export list.
    locToEEntry row col mod toks
      =if isJust exps 
          then let (_, e2) =break (\e->let (startLoc, endLoc) =getStartEndLoc toks e
                                       in startLoc<=(row,col) && (row,col)<=endLoc) (fromJust exps)
               in if e2==[] then Nothing else Just (head e2)
           else Nothing
 
       where exps = hsModExports mod     
 
    refactorInClientMod ents (modName@(PlainModule modName'), fileName)
      = do (inscps, exps, mod ,ts) <-parseSourceFile fileName          
           -- The following condition is too strong.         
           if findEnt ents (hsModDecls mod) || (intersect (map impEntToExpEnt ents) exps) /= []
            then  error $ "This entity can not be removed as it is used by the client module '"++modName'++"'!"
            else do let pnsToBeHided =findPns ents (hsModImports mod) 
                    if pnsToBeHided/=[]  
                     then do (mod', ((ts',m),_))<-runStateT (rmItemFromImport mod pnsToBeHided)
                                                   ((ts,unmodified),(fileName, initValue))
                             return ((fileName,m), (ts',mod'))
                     else return ((fileName,unmodified), (ts,mod)) 
    findEnt ents
        = (fromMaybe False).(applyTU (once_tdTU (failTU `adhocTU` worker)))
          where 
            worker (pnt::PNT)
             |isTopLevelPNT pnt && isJust (find (==(pntToEnt pnt)) ents) = Just True
            worker _ = Nothing

    impEntToExpEnt ((SN (UnQual s) loc1), e) = ((SN s loc1), e)
    impEntToExpEnt ((SN (Qual modName s) loc1),e) = ((SN s loc1), e)

    findPns ents 
        = (nub.ghead "findPns").(applyTU (full_tdTU (constTU [] `adhocTU` worker)))
          where 
            worker (pnt::PNT)
             |isTopLevelPNT pnt && isJust (find (==(pntToEnt pnt)) ents) = return [pNTtoPN pnt]
            worker _ = return []


 