
### Functions for use in shell scripts #########################################

errmsg() {
  echo >&2 "$*"
}

warning() {
  echo >&2 "Warning: $*"
}

abort() {
  errmsg "$*"
  exit 1
}

# Reading certificate/server attributes:
getattr() {
  egrep "^$1:" <"$2" | sed "s%^[a-zA-Z/.]*: *%%"
}

certauxfiles() {
  attrs="$1/cert.attr"
  echo "$certsDir$attrs"
  getattr "file(/[a-z]*)?" "$attrs"
}

# Check that a module exists:
checkmodule() {
  pfe pragmas "$1" >/dev/null || { echo "$1"; exit 1; }
}

# Check that a project exists:
checkproject() {
  [ -d hi ] ||
    abort "No pfe project in this directory. (Use pfesetup or pfe new.)"

  if pfeclient >/dev/null  2>&1 ; then
    pfe() {
     pfeclient "$@"
    }
  else
    warning "PFE server not found. Things can get slow..."
  fi
}

# Check that a certificate exists and set $attr:
checkcert() {
  attr="$certsDir$1/cert.attr"
  [ -r "$attr" ] ||
    abort "Didn't find certificate $1. (Use 'cert new' to create certificates.)"
}

# List of modules that a specified modules depends on
subgraph() {
  pfe modules $* | tail +2
}

# The list of source files that a specified module depends on
subgraphfiles() {
  pfe file `subgraph $*` | sed 's/^.*: //'
}

# Create a datestamp file when a certificate is marked valid/invalid
# Also save the list of source files that the certificate depends on
datestamp() {
  valid="${1-valid}"
  certpath=$2 # the path of a certificate directory
  module=$3 # the name of the module containing the certificate
  echo -n "Certificate marked $valid on "
  date | tee "$certpath/$valid"
  subgraphfiles $module >$certpath/srclist.txt
}

# Ask the user to confirm somthing (answer a yes-or-no question)
user_confirmation() {
  tty -s || echo >&2 "+-+ Yes No" # For pfebrowser
  read ans
  case "$ans" in
    [Yy]es) true ;;
    *) false
  esac
}
